package eu.pequignat;

import eu.pequignat.function.*;
import eu.pequignat.problem.pc.Consumer;
import eu.pequignat.problem.pc.Producer;
import eu.pequignat.problem.pc.QueuePC;
import eu.pequignat.runnable.MyCallable;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.time.Duration;
import java.time.Instant;
import java.util.*;
import java.util.concurrent.*;

import static eu.pequignat.constant.Constant.*;

/**
 * Hello world!
 *
 */
public class App 
{
    private static final Logger LOG = LoggerFactory.getLogger(App.class);


    public static void message(String tree, String message){
        LOG.info(String.format("%s> %s", tree ,message));
    }

    public static void launch(MyFunction myFunction){
        final String functionName= myFunction.getName();
        message(functionName, "BEGIN");
        Instant start = Instant.now();
        myFunction.run();
        Instant end = Instant.now();
        Duration timeElapsed = Duration.between(start, end);
        message(functionName, "END in ms: ".concat(String.valueOf(timeElapsed.toMillis())));
        message(functionName, "-----------------------------------------------");
    }


    /**
     * This function show how is the behaviour of a Thread.
     */
    public static void case1SimpleThreadRunnable(){
        message("case1SimpleThreadRunnable", "Just One Thread");
        launch(new SimpleThreadRunnable());
    }

    /**
     * This is not to be use as in really conditions. Just for this course
     */
    public static void case2MultiThreadingManuallyManaged(){
        launch(new MultiThreadingManuallyManaged());
    }

    /**
     * This methods fix the number of Thread to run in parallel
     */
    public static void case3FixedPoolOfThread(){
        launch(new MultiThreadingFixedPoolExecutor());
    }


    /**
     * This method use the Stream to make the parallelism
     */
    public static void case4MultiThreadingByStream(){
        int parallelism = Math.max(1, Runtime.getRuntime().availableProcessors() - 1);
        message("case4MultiThreadingByStream", "Use the default number of parallel:".concat(String.valueOf(parallelism)));
        launch(new MultiTreadingByStream());
    }

    /**
     * The number of parallelism to NB_THREAD_IN_PARALLEL.
     */
    public static void case5MultiThreadingByForkJoin(){
        message("case5MultiThreadingByForkJoin", "Use the number of parallel:".concat(String.valueOf(NB_THREAD_IN_PARALLEL)));
        launch(new MultiTreadingByForkJoinPool(NB_THREAD_IN_PARALLEL));
    }

    /**
     * Try to limit the number of parallelism to 2
     */
    public static void case5MultiThreadingByForkJoinLess(){
        int parallel = Math.min(LIMITED_NB_FORK_STREAM, Math.max(1, Runtime.getRuntime().availableProcessors() - 1));
        message("case5MultiThreadingByForkJoinLess", "Use the limited number of parallel:".concat(String.valueOf(parallel)));
        launch(new MultiTreadingByForkJoinPool(parallel));
    }

    /**
     * The classic demo of resolution of problem Consumer/Producer
     */
    public static void case6ConsumerProducerProblem(){
        message("case6ConsumerProducerProblem", "The classic demo of resolution of problem Consumer/Producer");
        //Use lambda expresson to instanciate a MyFunction
        MyFunction function = ()->{
            QueuePC queuePC = new QueuePC();
            Consumer consumer= new Consumer(queuePC);
            Producer producer=new Producer(queuePC);
            //Start all
            try {
                Thread consumerThread = new Thread(consumer);
                Thread producerThread = new Thread(producer);
                consumerThread.start();
                producerThread.start();
                for (int i=0;i< NB_LOOP; i++){
                    try {
                        LOG.info("SLEEP");
                        Thread.sleep(SLEEP_MS);
                    } catch (InterruptedException e) {
                        LOG.error(e.getLocalizedMessage(),e);
                    }
                }
                //Stop properly
                LOG.info("STOP");
                queuePC.stop();

                LOG.info("How to know when finished.");
                producerThread.join();
                consumerThread.join();
            } catch (InterruptedException e) {
                LOG.error(e.getLocalizedMessage(), e);
            }
        };
        launch(function);
    }


    public static void main( String[] args )
    {
        message("Main", "Starting...");
        /*
        A sample of use one Thread with a Runnable, in order to understand the behaviour
         */
        case1SimpleThreadRunnable();
        message("Main", "All tests below have the number of functions: ".concat(String.valueOf(NB_THREAD_IN_PARALLEL)));
        //Not recommaned implementations of Threads
        case2MultiThreadingManuallyManaged();
        //Better implamentation of fixedThread with Fixed Pool
        case3FixedPoolOfThread();

        //Use Stream to make parallelism and automatic wait the end of Thread
        case4MultiThreadingByStream();

        //Use same number of function that parallel number with ForkJoinPool
        case5MultiThreadingByForkJoin();

        //Limit the number of parallel ==> Take more time
        case5MultiThreadingByForkJoinLess();

        //The classic problem of Consumer/Producer
        case6ConsumerProducerProblem();

    }
}
